package de.hdm_stuttgart.mi.sd1.grep;

import java.io.BufferedReader;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.IOException;
import java.io.InputStreamReader;

import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;

/**
 * Entry class for poor man's UNIX grep. Implemented optional command line arguments:
 * 
 * <dl>
 *  <dt>-i</dt>
 *  <dd>Ignore case</dd>
 *  
 *  <dt>-l</dt>
 *  <dd>only show filenames containing matches rather than the matched lines themselves.</dd>
 * </dl>
 * 
 */
public class Grep {

	public class StreamAndMeta {

		public StreamAndMeta(final BufferedReader reader, final String streamName) {
			this.reader = reader;
			this.streamName = streamName;
		}

		public final BufferedReader reader;
		public final String streamName;

	}


	private static Logger log = LogManager.getLogger(Grep.class);

	final LineFilter lineFilter;
	final String searchString;

	final LineProcessor lineProcessor;

	/**
	 * if the -l option is present this value will be set to false showing just
	 * filenames rather than matching lines, false otherwise.
	 */
	public final boolean showMatchingData;

	final String[] filenames;

	public Grep(final String[] args) throws CommandLineException {

		boolean ignoreCaseTmp = false,
				showMatchingDataTmp = true; // Default values

		String searchStringTmp = null;

		final String[] maxFilenames = new String[args.length];// Possibly all command line arguments are just filenames

		int currentFilenameIndex = 0;

		for (int i = 0; i < args.length; i++) {
			final String arg = args[i];
			log.debug("command line argument '" + arg + "'");
			if (arg.startsWith("-")) {
				switch(arg) {
				case "-i":
					ignoreCaseTmp = true; 
					break;
				case "-l":
					showMatchingDataTmp = false;//Just show matching files, not the actual lines
					break;
				default:
					throw new CommandLineException("Invalid option '" + arg + "'", i);
				}
			} else {
				if (null == searchStringTmp) {
					searchStringTmp = arg;
				} else {
					maxFilenames[currentFilenameIndex++] = arg;
				}
			}
		}
		if (null == searchStringTmp) {
			throw new CommandLineException("No search string given", -1);
		}
		searchString = searchStringTmp;
		filenames = new String[currentFilenameIndex];
		log.debug("found " + currentFilenameIndex + " file(s):");
		for (int i = 0; i < currentFilenameIndex; i++) {
			filenames[i] = maxFilenames[i];
			log.debug("  filename '" + filenames[i]);
		}

		if (ignoreCaseTmp) {
			lineFilter = new CaseInsensitiveLineFilter(searchString);
		} else {
			lineFilter = new CaseSensitiveLineFilter(searchString);
		}
		showMatchingData = showMatchingDataTmp;

		lineProcessor = new LineProcessor(lineFilter);

		log.debug("ignoreCase=" + ignoreCaseTmp);
		log.debug("showMatchingData=" + showMatchingData);
	}

	static public String usage() {
		return "Usage: grep [-i] [-l] [filename1] ...";
	}

	/**
	 * Start processing input files or reading from standard input.
	 */
	public void process() {

		try {

			final StreamAndMeta streamsMeta[];
			if (0 == filenames.length) {
				streamsMeta = new StreamAndMeta[]{new StreamAndMeta(
						new BufferedReader (new InputStreamReader(System.in)),
						"(standard input)")};
			} else {
				streamsMeta = new StreamAndMeta[filenames.length];
				for (int i = 0; i < filenames.length; i++) {
					streamsMeta[i] = new StreamAndMeta(new BufferedReader(new FileReader(filenames[i])), filenames[i]);
				}
			}

			if (showMatchingData) {
				lineProcessor.process(streamsMeta);
			} else  {
				lineProcessor.findSingleMatch(streamsMeta);
			}

		} catch (final FileNotFoundException e) {
			System.err.println("File not found: " + e.getLocalizedMessage());
			System.exit(1);
		} catch (final IOException e) {
			System.err.println("Unable to open file:" + e.getLocalizedMessage());
			System.exit(1);
		}			

	}

	/**
	 * @param args Command line arguments guiding the desired matching matching process
	 * 
	 * This corresponds to the following command line usage:
	 * 
	 * grep [-i] [-l] searchString [file 1] [file 2] ...
	 * 
	 */
	public static void main(final String[] args) {

		try {
			final Grep grep = new Grep(args);
			grep.process();
		} catch (final CommandLineException e) {
			System.err.println(e.getMessage());
			System.err.println("Usage: grep [-i] [-l] searchString [file 1] [file 2] ...");
		}
	}
}