package de.hdm_stuttgart.mi.sd1;

/**
 * <p>Representing time periods measured in seconds by decomposing them into
 * seconds, minutes hours, days and weeks. See {@link #Timeperiod(int)}
 * for details.</p>
 */
public class Timeperiod {

    // Pre-caculated internal values
    static private final int
            SECONDS_PER_MINUTE = 60,                    // 60 minutes per hour
            SECONDS_PER_HOUR = 60 * SECONDS_PER_MINUTE, // 3600 seconds per hour
            SECONDS_PER_DAY = 24 * SECONDS_PER_HOUR,    // 86400 seconds per day
            SECONDS_PER_WEEK = 7 * SECONDS_PER_DAY;     // 604800 seconds per week

    // Tedious calculation, will happen only on-demand
    private String toStringValue = null;

    private String toStringImplement() {
        final int highestValue;

        if (0 < weeks) {
            highestValue = 5;       // weeks, days, hours, minutes and seconds
        } else if (0 < days) {
            highestValue = 4;       // days, hours, minutes and seconds
        } else if (0 < hours) {
            highestValue = 3;       // hours, minutes and seconds
        } else if (0 < minutes) {
            highestValue = 2;       // minutes and seconds
        } else  {
            highestValue = 1;       // only seconds, potentially zero as well
        }

        final StringBuffer buffer = new StringBuffer();
        switch (highestValue) {
            case 5: addSingularOrPlural(buffer, weeks, "week", ", ");
            case 4: addSingularOrPlural(buffer, days, "day", ", ");
            case 3: addSingularOrPlural(buffer, hours, "hour", ", ");
            case 2: addSingularOrPlural(buffer, minutes, "minute", ", ");
            case 1: addSingularOrPlural(buffer, seconds, "second", " and ");
        }
        return buffer.toString();
    }

    static private void addSingularOrPlural(final StringBuffer results, final int count,
                                            final String base, String separator) {

        if (0 < results.toString().length()) { // Result buffer not empty? Separator required.
            results.append(separator);
        }
        results.append(count).append(" ").append(base);
        if (1 != count) {
            results.append("s"); // 0 or plural, e.g. "3 seconds"
        }
    }

    /**
     * <p>Turning a given number of seconds into weeks, days, hours, minutes and seconds.
     *  * Example: 3729781 seconds are equivalent to the sum of 1 second, 3 minutes, 4 hours,
     *  * 1 day and 6 weeks.</p>
     *
     * @param seconds A time period provided in seconds.
     */
    public Timeperiod(int seconds) {

        weeks = seconds / SECONDS_PER_WEEK;
        seconds = seconds % SECONDS_PER_WEEK;        // remaining seconds without weeks

        days = seconds / SECONDS_PER_DAY;
        seconds %= SECONDS_PER_DAY;                  // remaining seconds without days

        hours = seconds / SECONDS_PER_HOUR;
        seconds %= SECONDS_PER_HOUR;                 // remaining seconds without minutes

        minutes = seconds / SECONDS_PER_MINUTE;
        this.seconds = seconds % SECONDS_PER_MINUTE; // remaining seconds
    }

    /**
     *  Copy a given instance.
     *
     * @param timeperiod Instance to be cloned.
     */
    public Timeperiod(final Timeperiod timeperiod) {
        seconds = timeperiod.seconds;
        minutes = timeperiod.minutes;
        hours = timeperiod.hours;
        days = timeperiod.days;
        weeks = timeperiod.weeks;
        toStringValue = timeperiod.toStringValue;
    }

    /**
     * <p>providing a human readable form. The output corresponding to e.g. 112223 seconds will read:</p>
     *
     * <pre>1 day, 7 hours, 10 minutes and 23 seconds</pre>
     *
     * <p>Notice the absence of a "0 weeks" clause: Output starts with the »highest« non-zero component.</p>
     *
     * @return Informal description e.g. "1 day, 7 hours, 10 minutes and 23 seconds".
     */
    @Override
    public String toString() {
        // Value calculation is tedious, thus doing it only on-demand.
        //
        if (null == toStringValue) { // Called for the first time, yet uninitialized?
            toStringValue = toStringImplement();
        }
        return toStringValue;
    }

    /**
     * Time period's seconds portion.
     */
    public final int seconds;
    /**
     * Time period's minutes portion.
     */
    public final int minutes;
    /**
     * Time period's hours portion.
     */
    public final int hours;
    /**
     * Time period's days portion.
     */
    public final int days;
    /**
     * Time period's weeksportion.
     */
    public final int weeks;
}