package de.hdm_stuttgart.sd1.math;

/**
 * <p>This class implements a subset of functions from class
 * {@link java.lang.Math} using power series expansions.</p>
 *
 */
public class Math {
  
  private static final double PI = java.lang.Math.PI;
  
  static int seriesLimit = 5;
  
  /**
   *  
   * @param seriesLimit The number of terms of a power series to be included.
   * 
   */
  public static void setSeriesLimit(int seriesLimit) {
    Math.seriesLimit = seriesLimit;
  }
  
  /**
   * <p>Approximating the natural exponential function by a finite
   * number of terms from the corresponding power series expansion.</p>
   *
   * \[ \begin{aligned}
   *   e^x ={} &amp; 1 + {x\over 1!} + {x^2\over 2!} + {x^3\over 3!} + \dots \\
   *       ={} &amp; \sum_{i = 0}^\infty {x^i\over i!}
   * \end{aligned} \]
   *
   * A power series implementation has to be finite since an
   * infinite number of terms requires infinite execution time.
   *
   * The number of terms to be considered can be set by {@link #setSeriesLimit(int)}}
   *
   * @param x The exponential's argument as in  \( e^x \)

   * @return The value \( e^x \) itself.
   */
  public static double exp(double x) {
    double currentTerm = 1.,  // the first (i == 0) term x^0/0!
        sum = currentTerm;    // initialize to the power series' first term
    
    for (int i = 1; i <= seriesLimit; i++) { // i = 0 has already been completed.
      currentTerm *= x / i;
      sum += currentTerm;
    }
    return sum;
  }
  
  /**
   * <p>Calculating the sine by means of the corresponding power series.</p>
   *
   \[
   \begin{aligned}
   \text{sin}(x) ={} &amp; x - {x^3\over 3!} + {x^5\over 5!} - {x^7\over 7!} + \dots \\
   ={} &amp; \sum_{i = 0}^\infty {(-1)^i  {x^{2i+1}\over (2i+1)!}}
   \end{aligned}
   \]
   *
   * @param x The sine's argument as in \( \text{sin}(x) \).
   *
   * @return The value \( \text{sin}(x) \) itself.
   */
  public static double sin(double x) {
    // Step 1: Normalize x to [-PI, +PI[
    final long countTimes2PI = (long) java.lang.Math.rint(x / 2 / PI);
    if (countTimes2PI != 0) {
      x -= 2 * PI * countTimes2PI;
    }
    
    // Step 2: Normalize x to [-PI/2, +PI/2]
    // Since sin(x) = sin (PI - x) we continue to normalize
    // having x in [-PI/2, +PI/2]
    if (PI/2 < x) {
      x = PI - x;
    } else if (x < -PI/2) {
      x = -x - PI;
    }

    // Step 3: Continue with business as usual
    
    double currentTerm = x,   // the first (i == 0) term x^1/1!
        sum = currentTerm;    // initialize to the power series' first term

    // Slightly incorrect if seriesLimit is even
    for (int i = 1; i <= seriesLimit; i += 2) { // Caution: Steps of two!!
      final double previousTerm =               // save the current term
          currentTerm *= x * x / (2 * i) / (2 * i + 1) ;
      currentTerm *= x * x / (2 * i + 2) / (2 * i + 3);
      sum += currentTerm - previousTerm;      //Summing up reordered
    }
    return sum;
  }}
