package de.hdm_stuttgart.mi.xmlstat;

import java.io.IOException;
import java.util.Map;
import java.util.stream.Collector;
import java.util.stream.Collectors;

import org.jdom2.Document;
import org.jdom2.Element;
import org.jdom2.JDOMException;
import org.jdom2.Namespace;
import org.jdom2.filter.ElementFilter;
import org.jdom2.input.SAXBuilder;
import org.jdom2.xpath.XPathExpression;
import org.jdom2.xpath.XPathFactory;

import sax.MySaxErrorHandler;

/**
 * Listing elements, namespaces and frequencies of arbitrary XML documents
 *
 */

public class ElementStatistics {
   
   public final Map<Namespace, ElementFrequency> elementsByNamespace;
   
   private final SAXBuilder builder = new SAXBuilder();
   private HtmlTree htmlOut = new HtmlTree("Namespaces and elements");


   /** Parse an Xml document providing a list of unique element names
    * among with their respective namespaces and
    * frequencies of appearance. 
    * 
    * @param 
    *   filename of the Xml source.
    * @throws IOException In case filename cannot be opened
    * @throws JDOMException Parsing error
    * 
    */
   public ElementStatistics(final String filename)
         throws JDOMException, IOException {
      builder.setErrorHandler(new MySaxErrorHandler(System.err));
      final Document doc = builder.build(getClass().getClassLoader().
            getResource(filename));
      
      final XPathExpression<Element> xpath = XPathFactory.instance().compile("//*",
            new ElementFilter());
      
      elementsByNamespace = 
      xpath.evaluate(doc).stream().collect(
            Collectors.groupingBy(
                  Element::getNamespace, 
                  Collector.of(
                        ElementFrequency::new,
                        ElementFrequency::add,
                        (left, right) -> {
                           left.merge(right); 
                           return left;
                        }
                  )
            )
      );
      
      elementsByNamespace.
      entrySet().
      stream().
      sorted(           // Sort by namespace URI
         (lEntry, rEntry) -> 
         lEntry.getKey().getURI().compareTo(rEntry.getKey().getURI())).
      forEach(htmlOut::appendNamespace);
      
      htmlOut.serialize(System.out);
   }
}